#define PI 3.1415926535897932384626433832795

// Tuto vzdálenost budeme považovat za nekonečno
#define FAR 50.0

vec4 quaternionMultiply(vec4 q, vec4 r)
{
    vec4 t;
    t.x = r.x*q.x - r.y*q.y - r.z*q.z - r.w*q.w;
    t.y = r.x*q.y + r.y*q.x - r.z*q.w + r.w*q.z;
    t.z = r.x*q.z + r.y*q.w + r.z*q.x - r.w*q.y;
    t.w = r.x*q.w - r.y*q.z + r.z*q.y + r.w*q.x;
    return t;
}

float lengthSquared(vec4 v)
{
    // dot(v, v) = v.x*v.x + v.y*v.y
    // + v.z*v.z + v.w*v.w = length(v)^2
    return dot(v, v);
}

float sdfMandelbrot(vec4 z, vec4 c)
{
    vec4 dz = vec4(1.0, 0.0, 0.0, 0.0);

    const int iterations = 32;

    for (int i = 0; i < iterations; i++)
    {
        dz = 2.0 * quaternionMultiply(z, dz);
        z = quaternionMultiply(z, z) + c;
        if (lengthSquared(z) > 256.0)
            break;
    }

    float distance = sqrt(
            lengthSquared(z) / lengthSquared(dz)
        ) * 0.5 * log(lengthSquared(z));

    return distance * 0.5;
}

float sdfSphere(vec3 center, float radius, vec3 p)
{
    return length(center - p) - radius;
}

float sceneSdf(vec3 p)
{
    float d = p.y;
    // Fraktál posuneme trochu nahoru
    vec3 fractalOffset = vec3(0.0, 1.3, 0.0);
    // SDF fraktálu
    d = min(d, sdfMandelbrot(vec4(p - fractalOffset, 0.0), vec4(-1.5,7,16.5,-6)/22.0));
    // Usekneme vše rovinou ve výšce y=1.3
    d = max(d, (p - fractalOffset).y);
    return d;
}

// Raymarchovací funkce
// Vstupy:
// origin - počátek paprsku
// direction - směr paprsku (vektor normalizovaný na velikost 1)
// Výstup: vzdálenost, kterou musí paprsek urazit, než narazí na kolizi
// nebo hodnota větší nebo rovnou než FAR, pokud do vzdálenosti FAR na
// žádnou kolizi nenarazil.
// Nezapomeňte smyčku ukončit, pokud hodnota SDF klesne pod nějaké epsilon
// nebo pokud se provede příliš mnoho iterací.
float raymarch(vec3 origin, vec3 direction)
{
    const float epsilon = 0.003;
    float marched = 0.0;
    int i;
    const int iterations = 256;
    for(i = 0; i < iterations; i++)
    {
        float dist = sceneSdf(origin + direction * marched);
        if(dist < epsilon)
            break;
        marched += dist;
        if(marched > FAR)
            break;
    }
    if(i >= iterations)
        return (FAR + 1.0);
    return marched;
}

vec3 computeNormal(vec3 p)
{
    const float epsilon = 0.001;
    float here = sceneSdf(p);
    return normalize(vec3(
        sceneSdf(p + epsilon * vec3(1.0, 0.0, 0.0)) - here,
        sceneSdf(p + epsilon * vec3(0.0, 1.0, 0.0)) - here,
        sceneSdf(p + epsilon * vec3(0.0, 0.0, 1.0)) - here
    ));
}

// forward: normalizovaný vektor dopředu
// uv: souřadnice pixelu v -1..1
// hfov: horizontální zorný úhel ve stupních
vec3 getRayDirection(vec3 forward, vec2 uv, float hfov)
{
    // Z "forward" vygenerujeme i vektory vpravo a nahoru, vše normalizované a navzájem kolmé
    vec3 right = normalize(cross(forward, vec3(0.0, 1.0, 0.0)));
    vec3 up = normalize(cross(right, forward));
    // Převedeme hfov na radiány
    hfov = hfov / 180.0 * PI;
    // Spočítáme vzdálenost roviny od kamery
    float dist = 1.0 / tan(hfov * 0.5);
    // Přeškálujeme "up" tak, aby byl zachován poměr stran
    up *= iResolution.y / iResolution.x;
    vec3 ray = forward * dist + right * uv.x + up * uv.y;
    return normalize(ray);
}

// Počítání Phongova osvětlovacího modelu
vec3 computeLight(vec3 colorDiffuse, vec3 colorSpecular, float a,
				  vec3 normal, vec3 view, vec3 lightDirection, vec3 lightIntensity)
{
    vec3 r = reflect(-lightDirection, normal);
    vec3 diffuse = lightIntensity * colorDiffuse * max(dot(normal, lightDirection), 0.0);
    vec3 specular = lightIntensity * colorSpecular * max(dot(normal, lightDirection), 0.0) *
        pow(max(dot(view, r), 0.0), a);
    return diffuse + specular;
}

vec3 getSky(vec3 direction)
{
	return mix(vec3(0.8, 0.9, 1.0), vec3(0.5, 0.7, 1.0), max(direction.y, 0.0));
}

vec3 getPixelColor(vec3 position, vec3 view, vec3 normal)
{
	// Zde jsme odstranili speciální připad pro zrcadlo
    
    vec3 surfaceColor = vec3(0.5);
    // Zem zabarvíme zeleně
    if(position.y < 0.03)
    {
    	surfaceColor = vec3(0.3, 0.5, 0.3);
    }
    
    vec3 surfaceSpecularColor = vec3(0.5);
    float surface_a = 8.0;
    vec3 light_dir = normalize(vec3(0.9, 0.8, 1.0));
    // Barva slunečního světla
    vec3 light_color = vec3(1.0, 0.9, 0.8);
    
    // Vystřelíme stínový paprsek
    bool shadow = raymarch(position + normal * 0.01, light_dir) < FAR;
    
    if(shadow)
    	light_color = vec3(0.0); // Jsme ve stínu, vynulujeme barvu slunečního světla
    
    // Spočítáme příchozí světlo ze slunce
    vec3 finalColor = computeLight(surfaceColor, surfaceSpecularColor, surface_a,
    	normal, view, light_dir, light_color);
    // Přičteme slabší ambientní světlo podle barvy oblohy
    finalColor += surfaceColor * getSky(normal) * 0.2;
    
    return finalColor;
}

// Pro místo na jednotkové kouli zadané
// pomocí úhlů vrátí jeho 3d souřadnice
// Vstup jsou úhly v radiánech
// První úhel je horizontální, druhý vertikální
vec3 anglesToPosition(float hor, float ver)
{
    return vec3(
        cos(ver) * sin(hor),
        sin(ver),
        cos(ver) * cos(hor)
    );
}

void mainImage(out vec4 fragColor, in vec2 fragCoord)
{
    // Souřadnice pixelu v rozsahu -1..1
    vec2 uv = fragCoord/iResolution.xy * 2.0 - 1.0;

    vec2 mouse = (iMouse.xy / iResolution.xy) * vec2(2.0 * PI, 0.5 * PI) + vec2(PI, 0.0);
    vec3 cameraLookAt = vec3(0.0, 1.2, 0.0);
    vec3 cameraPosition = cameraLookAt + anglesToPosition(mouse.x, mouse.y) * 2.5;
    
    vec3 rayDirection = getRayDirection(
        normalize(cameraLookAt - cameraPosition),
        uv, 80.0);

    float depth = raymarch(cameraPosition, rayDirection);
    
    // Souřadnice dopadu paprsku
    vec3 hit = cameraPosition + rayDirection * depth;
    // Normála povrchu v bodě dopadu
    vec3 normal = computeNormal(hit);
    
    vec3 color = getPixelColor(hit, -rayDirection, normal);
    
    if(depth > FAR)
    {
        color = getSky(rayDirection);
    }
    
    // Výstup a gamma korekce
    fragColor = vec4(pow(color, vec3(1.0/2.2)), 1.0);
}
